@@ -1148,6 +1148,15 @@ sub singletest_postcheck {
             }
         }
     }
+
+    if($checktests) {
+        loadtest("${TESTDIR}/test${testnum}");  # load the raw original data
+        if(checktest()) {
+            logmsg " $testnum: postcheck FAILED: issue(s) found in test data\n";
+            return -1;
+        }
+    }
+
     return 0;
 }
 
